###
# not part of the interface!  used by subsidiary instruments
default_mode_names = {
    'imaging'           : 'Imaging',
    'rampfilter'        : 'Ramp Filter',
    'spectroscopicacq'  : 'Spectroscopy Target Acquisition',
    'spectroscopic'     : 'Spectroscopy',                       # BUG: why different?
    'targetacquisition' : 'Imaging Target Acquisition',
}

default_short_mode_names =  {
    'imaging'           : 'im',
    'rampfilter'        : 'rf',
    'spectroscopicacq'  : 'sa',
    'spectroscopy'      : 'sp',
    'targetacquisition' : 'ta',
}

