#
# automatically generated by tools/form_field_table/gen_form_fields.py
#
field_list = [
    'AirglowStandard',
    'EarthshineMag',
    'EarthshineMult',
    'EarthshineSpec',
    'EarthshineStandard',
    'Mirror',
    'SNR',
    'Time',
    'ZodiDate',
    'ZodiDay',
    'ZodiDec',
    'ZodiMag',
    'ZodiMonth',
    'ZodiMult',
    'ZodiRA',
    'ZodiSpec',
    'ZodiStandard',
    'ZodiSun',
    'ZodiSunAttribute',
    'cosaperture0',
    'extractionRegionCircle',
    'extractionRegionExtendedCircle',
    'extractionRegionExtendedUserCircle',
    'extractionRegionExtendedUserCircle_min',
    'extractionRegionExtendedUserCircle_max',
    'extractionRegionUserCircle',
    'extractionRegionUserCircle_min',
    'extractionRegionUserCircle_max',
    'extractionRegionVariableCircle',
    'extractionRegionSquare',
    'extractionRegionExtendedSquare',
    'fband',
    'fbandpass',
    'fIndex',
    'fIsLambda',
    'fL1_center',
    'fL1_flux',
    'fL1_fwhm',
    'fL2_center',
    'fL2_flux',
    'fL2_fwhm',
    'fL3_center',
    'fL3_flux',
    'fL3_fwhm',
    'fOtherFile',
    'fRedshift',
    'fSpectrumCK',
    'fSpectrumPickles',
    'fStellar',
    'fUploadFile',
    'fbbtemp',
    'fbpgsfile',
    'fcalfile',
    'fphoenixfile',
    'fdiameter',
    'febmvtype',
    'febv',
    'fextinctiontype',
    'rn_flux_lambda',
    'rn_flux_lambda_units',
    'rn_flux_bandpass',
    'rn_flux_bandpass_units',
    'rn_lambda',
    'rn_lambda_units',
    'fftype_filters',
    'filter.ubvri',
    'filter.sloan',
    'filter.acs',
    'filter.nicmos',
    'filter.wfc3UVIS',
    'filter.wfc3IR',
    'fftype',
    'fnonstellar',
    'fsorigin',
    'fsourceType',
    'instrument',
    'science_mode',
    'simmode',
    'xRegionExtendedType',
    'xRegionType',
]
